#-*- coding:utf-8 -*-

#    Pybik -- A 3 dimensional magic cube game.
#    Copyright (C) 2009  B. Clausius
#    Copyright (C) 2003  John Darrington
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License,  or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os

import gobject
import gtk
from gtk import gdk
import gio
import pango

import glarea
import drwBlock
import textures
import widget_set
import select_


BOX_PADDING = 10

N_ = lambda string: string


RDRAG_COLOUR = 0
RDRAG_FILELIST = 1

SWATCH_WIDTH  = 256
SWATCH_HEIGHT = 256
SWATCH_WIDTH_SMALL  = 64
SWATCH_HEIGHT_SMALL = 64
SWATCH_WIDTH_SMALLEST = 32
SWATCH_HEIGHT_SMALLEST = 32

UNDEFINED = 0
COLORED = 1
IMAGED = 2

TILED = 0
MOSAIC = 1

#class FacePreferences (object):
#    def __init__(self, color, pattern, filename, distr):
#        self.color = color
#        try:
#            self.gdkcolor = gdk.color_parse (color)
#        except ValueError:
#            print "colour %s not in database\n" % colour
#        self.pattern = pattern
#        self.filename = filename
#        self.distr = distr

class Preferences (object):
    def __init__(self):
        self.default_colors = ['#a81407','#d94b1c','#f0c829','#e3e3e3','#1d6311','#00275e']
        self.gdk_colors = []
        #self.faces = []
        for i,colour in enumerate(self.default_colors):
            #self.faces.append(FacePreferences(colour, -1, '', MOSAIC))
            try:
                self.gdk_colors.append(gdk.color_parse(colour))
            except ValueError:
                print "colour %s not parseabel" % colour
                self.gdk_colors.append(None)
        
        self.liststore_faces = gtk.ListStore(
                        gobject.TYPE_INT,       # 0 FaceNum
                        gobject.TYPE_STRING,    # 1 FaceName
                        gdk.Pixbuf,             # 2 Swatch
                        gdk.Color,              # 3 Color
                        gobject.TYPE_INT,       # 4 PatternNum
                        gobject.TYPE_STRING,    # 5 ImageFilename
                        gdk.Pixbuf,             # 6 ImagePixbuf
                        gobject.TYPE_INT,       # 7 DistrMode
                        gobject.TYPE_INT,       # 8 FaceType COLORED, IMAGED
                        gobject.TYPE_INT,       # 9 texname
                    )
        face_names = [_('Up'), _('Down'), _('Left'), _('Right'), _('Front'), _('Back')]
        for i in xrange(6):
            #FIXME
            self.liststore_faces.append(
                                (i, face_names[i], None,
                                self.gdk_colors[i],
                                -1,
                                None, None,
                                MOSAIC,
                                COLORED,
                                -1
                                ))
    
    def set_all(self, i):
        color = self.get_color(i)
        drwBlock.face_rendering_set (i, red=color.red / 65535.0,
                                        green=color.green / 65535.0,
                                        blue=color.blue / 65535.0,
                                        facetype=self.get_facetype(i),
                                        texname=self.get_texname(i),
                                        distr=self.get_distr(i))
    def get_color(self, i):         return self.liststore_faces[i][3]
    def set_color(self, i, color):
        self.liststore_faces[i][3] = color
        red   = color.red   / 65535.0
        green = color.green / 65535.0
        blue  = color.blue  / 65535.0
        #drwBlock.face_rendering_set_swatch_colour (i, red, green, blue)
        self.set_facetype(i, COLORED)
        drwBlock.face_rendering_set (i, red=red, green=green, blue=blue, facetype=COLORED, texname=-1)
    def get_pattern(self, i):         return self.liststore_faces[i][4]
    def set_pattern(self, i, pattern):
        self.liststore_faces[i][4] = pattern
        if pattern >= 0:
            texName = textures.stock_pattern[pattern].texName
        else:
            texName = -1
        #drwBlock.face_rendering_select_texture(i, pattern, texName)
        self.set_facetype(i, COLORED)
        drwBlock.face_rendering_set(i, texname=texName, facetype=COLORED)
        set_cube_rendering()
    def get_image_filename(self, i):    return self.liststore_faces[i][5]
    def get_image(self, i):         return self.liststore_faces[i][6]
    def get_texname(self, i):       return self.liststore_faces[i][9]
    def set_image_filename(self, i, image_filename):
        self.liststore_faces[i][5] = image_filename
        try:
            pixbuf = create_pixbuf_from_file (image_filename)
        except Exception, err:
            print err
            #FIXME
            #widget_set.error_dialog (fs, err.message)
            return False
        try:
            texName = textures.create_pattern_from_pixbuf (pixbuf)
        except Exception, err:
            print err
            #FIXME
            #widget_set.error_dialog (fs, _("Cannot create image from file %s: %s"),  filename, str(err))
            return False
        self.liststore_faces[i][6] = pixbuf
        self.liststore_faces[i][9] = texName
        #drwBlock.face_rendering_set_swatch_image(i, texName)
        self.set_facetype(i, IMAGED)
        drwBlock.face_rendering_set(i, texname=texName, facetype=IMAGED)
    def get_distr(self, i):     return self.liststore_faces[i][7]
    def set_distr(self, i, distr):
        self.liststore_faces[i][7] = distr
        #drwBlock.face_rendering_set_rendering_distr(i, distr)
        drwBlock.face_rendering_set(i, distr=distr)
    def get_facetype(self, i):  return self.liststore_faces[i][8]
    def set_facetype(self, i, facetype):
        self.liststore_faces[i][8] = facetype
        #drwBlock.face_rendering_set_rendering_distr(i, distr)

the_prefs = Preferences()

#@debug
def set_the_colours():
    for i,face in enumerate(the_prefs.liststore_faces):
        the_prefs.set_all(i)

def create_pixbuf_from_file (filename):
    if DEBUG:
        print 'create_pixbuf_from_file:', filename
    unscaled_pixbuf =  gdk.pixbuf_new_from_file(filename)
    
    # We must scale the image,  because Mesa/OpenGL insists on it being of size
    # 2^n ( where n is integer )
    pixbuf = unscaled_pixbuf.scale_simple(SWATCH_WIDTH, SWATCH_HEIGHT, gdk.INTERP_NEAREST)
    return pixbuf


@debug
def open_ColorSelectDialog(toplevel):
    if ColorSelectDialog.this_dlg:
        ColorSelectDialog.this_dlg.dialog.show_all()
    else:
        ColorSelectDialog(toplevel)

class ColorSelectDialog (object):
    this_dlg = None
    
    def __init__(self, toplevel):
        assert self.this_dlg == None
        ColorSelectDialog.this_dlg = self
        
        self.dialog = toplevel.get_object('dialog_edit_colors')
        self.dialog.connect ("response", lambda w,r: w.hide())
        self.dialog.connect ("delete-event", lambda *a: True)
        
        set_cube_rendering()
        
        self.active_face = None
        self.create_face_selector(toplevel)
        
        self.button_color = toplevel.get_object('colorbutton1')
        self.button_color.connect ("color-set", self.on_button_color_set)
        
        self.create_pattern_selector(toplevel)
        
        self.button_image = toplevel.get_object('filechooserbutton_image')
        self.button_image.connect ("file-set", self.on_button_image_file_set)
        self.button_image_tile = toplevel.get_object('radiobutton_tile')
        self.button_image_mosaic = toplevel.get_object('radiobutton_mosaic')
        self.button_image_tile.connect ("toggled", self.on_button_image_distr_toggled, TILED)
        self.button_image_mosaic.connect ("toggled", self.on_button_image_distr_toggled, MOSAIC)
        self.button_image_group = gtk.RadioButton.get_group (self.button_image_mosaic)
        
        self.docb = True
        self.dialog.show_all()

    def create_face_selector(self, toplevel):
        target = [
                    ( "text/uri-list",  0,  RDRAG_FILELIST ),
                    ( "application/x-color",  0,  RDRAG_COLOUR ),
                ]
            
        treeview_faces = toplevel.get_object('treeview_faces')
        treeview_faces.set_model(the_prefs.liststore_faces)
        
        column = gtk.TreeViewColumn(_('Face'), gtk.CellRendererPixbuf(), pixbuf=2)
        column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column.set_fixed_width(SWATCH_WIDTH_SMALLEST)
        treeview_faces.append_column(column)
        column = gtk.TreeViewColumn(_('Face'), gtk.CellRendererText(), text=1)
        column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column.set_fixed_width(50)
        treeview_faces.append_column(column)
        treeview_faces.connect ("cursor-changed", self.on_treeview_faces_cursor_changed)
        
    def create_pattern_selector(self, toplevel):
        texture_tip = 0

        button_box = toplevel.get_object('hbuttonbox1')
        self.combobox_pattern = toplevel.get_object('combobox_pattern')
        cr1 = gtk.CellRendererPixbuf()
        self.combobox_pattern.pack_end(cr1)
        cr2 = gtk.CellRendererText()
        #cr2.props.xalign = 0.5
        self.combobox_pattern.pack_end(cr2)
        liststore = gtk.ListStore(gobject.TYPE_INT, gobject.TYPE_INT, gdk.Pixbuf, gobject.TYPE_STRING)
        self.combobox_pattern.set_model(liststore)
        self.combobox_pattern.set_attributes(cr1, pixbuf=2)
        self.combobox_pattern.set_attributes(cr2, text=3)
        self.combobox_pattern.set_column_span_column(1)
        self.combobox_pattern.connect ("changed", self.on_combobox_pattern_changed)
        self.combobox_pattern.set_tooltip_text ("Click here to use a pattern on the cube surface")
        
        data = ''.join(chr(255)*8 for c in textures.stock_pattern[0].data)
        data = [((ord(c)>>(_i%8))&1)*255 for _i,c in enumerate(data)]
        data = ''.join(chr(c)*3+chr(255) for c in data)
        pixbuf = gdk.pixbuf_new_from_data(data, gdk.COLORSPACE_RGB, True, 8, 
                                textures.checkImageWidth, textures.checkImageHeight, 
                                textures.checkImageWidth*4)
        liststore.append((-1, 3, None, _('plain')))
        
        for i in xrange(6):
            data = ''.join(chr(c)*8 for c in textures.stock_pattern[i].data)
            data = [((ord(c)>>(_i%8))&1)*255 for _i,c in enumerate(data)]
            data = ''.join(chr(c)*3+chr(255) for c in data)
            pixbuf = gdk.pixbuf_new_from_data(data, gdk.COLORSPACE_RGB, True, 8, 
                                    textures.checkImageWidth, textures.checkImageHeight, 
                                    textures.checkImageWidth*4)
            liststore.append((i, 1, pixbuf, ''))
            
        
    @debug
    def on_treeview_faces_cursor_changed(self, widget):
        path = widget.get_cursor()[0]
        assert path
        self.active_face = widget.get_model()[path]
        #( face_num,
        #  face_name,
        #  swatch_pixbuf,
        #  color,
        #  pattern_num,
        #  image_file,
        #  image_pixbuf,
        #  distr_mode,
        #  face_type) = self.active_face
        #assert path[0] == face_num
        face_num = path[0]
        #print face_num, pattern_num
        
        self.docb = False
        self.button_color.set_color(the_prefs.get_color(face_num))
        self.combobox_pattern.set_active(the_prefs.get_pattern(face_num)+1)
        image_file = the_prefs.get_image_filename(face_num)
        if image_file:
            self.button_image.set_filename(image_file)
        else:
            self.button_image.unselect_all()
        distr_mode = the_prefs.get_distr(face_num)
        self.button_image_tile.set_active (distr_mode==TILED)
        self.button_image_mosaic.set_active (distr_mode==MOSAIC)
        self.set_distr_control_sensitivity(None, face_num)
        self.docb = True


    # This function sets the currently selected swatch to
    # the colour on the clr_sel widget
    def on_button_color_set (self, button):
        if self.docb:
            color = button.get_color()
            the_prefs.set_color(self.active_face[0], color)
            self.set_distr_control_sensitivity (None, self.active_face[0])
            set_cube_rendering()


    @debug
    def on_combobox_pattern_changed(self, widget):
        if self.docb:
            active_iter = widget.get_active_iter()
            if active_iter:
                pattern = widget.get_model()[active_iter][0]
                the_prefs.set_pattern(self.active_face[0], pattern)


    def on_button_image_file_set(self, widget):
        if self.docb:
            filename = widget.get_filename()
            the_prefs.set_image_filename(self.active_face[0], filename)
            self.set_distr_control_sensitivity (None, -1)
            set_cube_rendering()

    # This callback sets the distribution mode of the image
    def on_button_image_distr_toggled (self, b, distr):
        if self.docb:
            if  not gtk.ToggleButton.get_active (b):
                return
            the_prefs.set_distr(self.active_face[0], distr)
            glarea.postRedisplay()


    def get_pixbuf_for_swatch(self, i, size):
        #if drwBlock.face_rendering_is_rendering_type_COLORED(i):
        if the_prefs.get_facetype(i) == COLORED:
            #FIXME
            return None

        #elif drwBlock.face_rendering_is_rendering_type_IMAGED(i):
        elif the_prefs.get_facetype(i) == IMAGED:
            width, height = size,size
            pixbuf = the_prefs.get_image(i)
            assert pixbuf
            scaled_pixbuf = pixbuf.scale_simple(width, height, gdk.INTERP_NEAREST)
            assert scaled_pixbuf
            return scaled_pixbuf

        else:
            #FIXME
            #assert False, 'not_reached'
            return None


    def redraw_swatches(self):
        for i in xrange(6):
            #self.draw_swatch (self.swatches[i], None, i)
            #FIXME update pixbufs in liststore
            pass


    # Set the sensitivity of the TILED/MOSAIC buttons
    def set_distr_control_sensitivity (self, unused_widget, swatch):
        if swatch == -1:
            swatch = self.active_face[0]

        #if drwBlock.face_rendering_is_rendering_type_IMAGED(swatch):
        b = the_prefs.get_facetype(swatch) == IMAGED 
        for button in self.button_image_group:
            button.set_sensitive(b)

        return False


@debug
def drag_data_received (unused_widget,
            unused_dc, x, y, selection_data,
            info, unused_t, swatch=-1, color_select_dialog=None):
    # The special value -1 means that this callback must
    # use pickPolygons to find out which face is to be updated
    #FIXME Redefining name ... ; anderen Namen verwenden
    if swatch == -1:
        fs = select_.pickPolygons (x,  y)
        if not fs:  # If nothing was actually pointed to,  then exit
            return
        swatch = fs.face
        immediate_mode = True
    else:
        immediate_mode = False

    if info == RDRAG_COLOUR:
        #FIXME lenght and format available in PyGTK 2.14 and above.
        #if selection_data.length < 0:
        #    return
        #if (selection_data.format != 16) or (selection_data.length != 8):
        #    print "Received invalid color data"
        #    return
        if len(selection_data.data) != 8:
            print "Received invalid color data"
            return
        
        #vals = selection_data.data
        vals = [ord(selection_data.data[0]) | ord(selection_data.data[1]) << 8,
                ord(selection_data.data[2]) | ord(selection_data.data[3]) << 8,
                ord(selection_data.data[4]) | ord(selection_data.data[5]) << 8]
        colour = gdk.Color()
        if DEBUG:
            print (colour.red, vals)
        colour.red = vals[0]
        colour.green = vals[1]
        colour.blue = vals[2]

        if immediate_mode:
            #drwBlock.face_rendering_set_swatch_colour(swatch,
            #                                                vals[0] / 65535.0,
            #                                                vals[1] / 65535.0,
            #                                                vals[2] / 65535.0)
            drwBlock.face_rendering_set (swatch, 
                                        red=vals[0] / 65535.0,
                                        green=vals[1] / 65535.0,
                                        blue=vals[2] / 65535.0,
                                        facetype=COLORED)

        if ColorSelectDialog.this_dlg:
            the_prefs.set_color(swatch, color)         
            ColorSelectDialog.this_dlg.set_distr_control_sensitivity (None, swatch)
            ColorSelectDialog.this_dlg.redraw_swatches()
            set_cube_rendering()

    elif info == RDRAG_FILELIST:
        start = selection_data.data.split("\r\n")
        for s in start:
            if s == "":
                continue
            
            fileobj = gio.file_parse_name(s)
            filename = gio.file_parse_name(s).get_path()
            if not fileobj.query_exists():
                print 'filename "%s" not found.' % filename
                continue
            
            if immediate_mode:
                the_prefs.set_image_filename(swatch, filename)

            if ColorSelectDialog.this_dlg:
                the_prefs.set_image_filename(swatch, filename)
                ColorSelectDialog.this_dlg.set_distr_control_sensitivity (None, swatch_)
                ColorSelectDialog.this_dlg.redraw_swatches()

            # For now,  just use the first one.
            # Later,  we'll add some method for disambiguating multiple files
            break

        glarea.postRedisplay ()
    # Ignore all others


# Set the colours of the cube,  from the colour swatch
def set_cube_rendering ():
#    for i in xrange(6):
#        drwBlock.set_face_rendering (i)
    glarea.postRedisplay ()

