#-*- coding:utf-8 -*-

#    Pybik -- A 3 dimensional magic cube game.
#    Copyright (C) 2009  B. Clausius
#    Copyright (C) 1998,  2003  John Darrington
#                  2004  John Darrington,  Dale Mellor
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License,  or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import txfm
import cube
import drwBlock
import glarea
import select_
import widget_set

import quarternion


import move_queue_m
import glarea_common

import gtk.keysyms
import gobject


PLAY_TOOLBAR_BACK = 0x0b << 0
PLAY_TOOLBAR_STOP = 0x01 << 2
PLAY_TOOLBAR_PLAY = 0x03 << 4


GNUBIK_Left = gtk.keysyms.Left
GNUBIK_Right = gtk.keysyms.Right
GNUBIK_Up  = gtk.keysyms.Up
GNUBIK_Down = gtk.keysyms.Down


abort_requested = 0


# picture rate in ms. 40ms = 25 frames per sec
picture_rate = 40
animation_in_progress = 0
# this is the number of frames drawn when a slice of the cube rotates 90 degrees
frameQty = 2


move_queue = None


# The move taking place NOW
class _Current_Movement:
    def __init__(self):
        self.move_data = move_queue_m.MoveData()
        self.stop_requested = 0
current_movement = _Current_Movement


#  Toolbar callback. The stop will happen when the current animation
#  completes.
#@debug
def request_stop (unused_widget):
    current_movement.stop_requested = 1


# The move that will take place when the mouse is clicked
pending_movement = move_queue_m.MoveData(-1, -1, -1)


def is_animating ():
    return animation_in_progress


# Rotate cube about axis (screen relative)
def rotate_cube (axis, dir):
    # how many degrees to turn the cube with each hit
    step = 2.0
    rot = quarternion.Quarternion()
    #vector v
    v = txfm.Vector()

    if dir:
        step = -step

    if axis == 0:
        v[1] = v[2] = 0
        v[0] = 1
    elif axis == 1:
        v[0] = v[2] = 0
        v[1] = 1
    elif axis == 2:
        v[0] = v[1] = 0
        v[2] = 1

    quarternion.quarternion_from_rotation (rot,  v,  step)
    quarternion.quarternion_pre_mult (glarea_common.cube_view, rot)
    glarea.postRedisplay ()



# orientate the whole cube with the arrow keys
def do_keysyms (keysym, shifted):
    if keysym == GNUBIK_Right:
        if shifted:
            dir = 0
            axis = 2
        else:
            dir =1
            axis = 1
    elif keysym == GNUBIK_Left:
        if shifted:
            dir = 1
            axis = 2
        else:
            dir =0
            axis = 1
    elif keysym == GNUBIK_Up:
        axis = 0
        dir = 0
    elif keysym == GNUBIK_Down:
        axis = 0
        dir = 1
    else:
        return

    rotate_cube (axis,  dir)
    return True


# return the turn direction,  based upon the turn axis vector
#@debug
def turnDir (vector):
    # This is a horrendous kludge.  It works only because we know that
    # vector is arithemtically very simple

    return vector[0] + vector[1] + vector[2]  >  0




# Convert a vector to a axis number.  Obviously this assumes the vector
#is orthogonal to the frame of reference ( another nasty kludge ).
def vector2axis(vector):
    return (0 if vector[0] != 0 else
            1 if vector[1] != 0 else
            2)




turn_axesX  = [1, 0, 0, 0]
turn_axes_X = [-1, 0, 0, 0]
turn_axesY  = [0, 1, 0, 0]
turn_axes_Y = [0, -1, 0, 0]
turn_axesZ  = [0, 0, 1, 0]
turn_axes_Z = [0, 0, -1, 0]
turn_axes = [
                [turn_axes_Y, turn_axesX,  turn_axesY,  turn_axes_X],
                [turn_axesY,  turn_axesX,  turn_axes_Y, turn_axes_X],
                [turn_axesZ,  turn_axesX,  turn_axes_Z, turn_axes_X],
                [turn_axes_Z, turn_axesX,  turn_axesZ,  turn_axes_X],
                [turn_axes_Y, turn_axes_Z, turn_axesY,  turn_axesZ],
                [turn_axesY,  turn_axes_Z, turn_axes_Y, turn_axesZ],
            ]

# Determine the axis about which to rotate the slice,  from the objects
# selected by the cursor position
#@debug
def getTurnAxis (items):
    # Each edge (quadrant) on a block represents a different axis

    # Fetch the selected block's transformation from its original
    # orientation
    t = cube.get_block_transform(items.block)

    # Select the untransformed vector for the selected edge
    vector = [turn_axes[items.face][items.quadrant][i] for i in xrange(4)]

    # transform it,  so that we go the right way
    vector = txfm.transform (t, vector)
    
    return vector


# handle mouse clicks
def mouse (button, application):
    # Don't let a user make a move,  whilst one is already in progress,
    # otherwise the cube falls to bits.
    if animation_in_progress:
        return

    if button == 1:
        # Make a move
        if select_.itemIsSelected ():
            request_rotation (pending_movement, application)
        
        glarea.postRedisplay ()





# Toolbar callback. Use stop_requested to ensure that only one step occurs.

#@debug
def request_back (unused_widget, application):
    if not animation_in_progress and move_queue and move_queue.retard():
        current_movement.stop_requested = 1

        next_move = move_queue.current().data.copy()
        next_move.dir = not next_move.dir

        animate_rotation (next_move, application)



# Internal functionality for the following two functions.

#@debug
def _request_play (one_move, application):
    if not animation_in_progress and move_queue and move_queue.current():
        current_movement.stop_requested = one_move

        next_move = move_queue.current().data.copy()
        move_queue.advance()

        animate_rotation (next_move, application)


# Toolbar `play' callback.
#@debug
def request_play (unused_widget, application):
    _request_play (0, application)

# Toolbar `forward' callback.
#@debug
def request_forward (unused_widget, application):
    _request_play (1, application)



# The usual rotation request,  called when user uses mouse to rotate the cube by
# hand. The move is put into the `current' place,  and all the moves in the
# move_queue afterwards are zapped.

#@debug
def request_rotation (data, application):
    global move_queue
    if move_queue == None:
        move_queue = move_queue_m.MoveQueue()
    #print move_queue

    move_queue.push_current(data)

    request_play (None, application)




# The rotation request called from script-fu. In this case the moves are
# appended to the move_queue,  so the script can race ahead and work out the
# moves,  leaving the graphics to catch up later.

#@debug
def request_delayed_rotation (data):
    global move_queue
    if move_queue == None:
        move_queue = move_queue_m.MoveQueue()

    move_queue.push(data)




# Script-fu calls this function on startup,  and then the above function as each
# move is computed. This way,  the moves computed are appended to the move_queue
# from the current point forwards (and forward data that was on the queue is
# discarded).

#@debug
def request_truncate_move_queue ():
    if move_queue:
        move_queue.truncate()



# Toolbar callback.

#@debug
def request_mark_move_queue (unused_widget):
    if move_queue:
        move_queue.mark_current()




# Toolbar callback. All the moves take place on the cube,  not via the animation
# routines,  so the effect is to unwind the queue instantaneously.

#@debug
def request_queue_rewind (unused_widget, application):
    if not move_queue:
        return

    while True:
        mark_result = move_queue.retard()
        if mark_result == 0:
            break

        move_data = move_queue.current().data
        if move_data != None:
            cube.rotate_back(move_data)
        
        if mark_result != 1:
            break

    glarea.postRedisplay ()
    ##toolbar_state = 0
    ##if move_queue.progress().current:   toolbar_state |= PLAY_TOOLBAR_BACK
    ##if move_queue.current():            toolbar_state |= PLAY_TOOLBAR_PLAY
    #widget_set.set_toolbar_state (toolbar_state)
    update_ui(application)



# Script callback.

#@debug
def request_fast_forward (application):
    if not move_queue:
        return

    move_data = move_queue.current()
    while move_data != None:
        blocks = cube.identify_blocks_2 (move_data.slice, move_data.axis)
        cube.rotate_slice ( 3 if move_data.dir==0 else 1,
                            blocks)
        
        move_queue.advance()
        move_data = move_queue.current()

    glarea.postRedisplay ()
    #widget_set.set_toolbar_state (PLAY_TOOLBAR_BACK)
    update_ui(application)



# Does exactly what it says on the tin :-)

#@debug
def animate_rotation (data, application):
    assert isinstance(data, move_queue_m.MoveData), str(data)
    current_movement.move_data = data.copy()

    drwBlock.start_animate(data)

    global animation_in_progress
    animation_in_progress = 1

    widget_set.set_toolbar_state (PLAY_TOOLBAR_STOP)

    #print 'picture_rate',picture_rate
    assert type(application)==widget_set.MainWindow, application
    gobject.timeout_add (picture_rate, on_animate, current_movement, application)


def abort_animation ():
    global abort_requested
    abort_requested = 1


# a timeout  calls this func,  to animate the cube
#@debug
def on_animate(data, application):
    global abort_requested
    #assert isinstance(data, move_queue_m.MoveData), repr(data)
    md = data
    #assert isinstance(md.move_data, move_queue_m.MoveData), str(md.move_data)

    # how many degrees motion per frame
    increment = 90.0 / (frameQty + 1)

    # decrement the current angle
    unfinished = drwBlock.step_animate(increment)

    # and redraw it
    glarea.postRedisplay ()

    if unfinished and not abort_requested:
        # call this timeout again
        gobject.timeout_add (picture_rate, on_animate, data, application)
    else:
        # we have finished the animation sequence now
        drwBlock.end_animate(md.move_data)

        global animation_in_progress
        animation_in_progress = 0
        
        select_.updateSelection ()
        
        #toolbar_state = 0
        #if move_queue.progress().current:   toolbar_state |= PLAY_TOOLBAR_BACK
        #if move_queue.current():            toolbar_state |= PLAY_TOOLBAR_PLAY
        #widget_set.set_toolbar_state (toolbar_state)
        update_ui(application)

        if abort_requested:
            abort_requested = 0

        # If there are more animations in the queue,  go again unless a stop has
        # been requested.

        if md.stop_requested:
            md.stop_requested = 0

        elif move_queue.current():
            md.move_data = move_queue.current().data.copy()
            move_queue.advance()
            animate_rotation (md.move_data, application)
        
    return False

def update_ui(application):
    toolbar_state = 0
    if move_queue.progress().current:   toolbar_state |= PLAY_TOOLBAR_BACK
    if move_queue.current():            toolbar_state |= PLAY_TOOLBAR_PLAY
    widget_set.set_toolbar_state (toolbar_state)
    
    application.update_statusbar()
    
    status = cube.cube_status_check()
    if cube.NOT_SOLVED != status:
        application.declare_win (status)

