#-*- coding:utf-8 -*-

#    Pybik -- A 3 dimensional magic cube game.
#    Copyright (C) 2009  B. Clausius
#    Copyright (C) 2003, 2004  John Darrington
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License,  or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import config
import menus
import colour_sel
import plugin_hooks
import move_queue_m
import ui

import gobject
import gtk
from gtk import gdk
import gtk.glade
from gtk import gtkgl
from gtk import gdkgl

import drwBlock
import glarea

N_ = lambda string: string

import cube


main_application_window = None
play_toolbar = None
play_button_list = None

class MainWindow(object):
    def __init__(self, argv):
        global main_application_window, play_toolbar
        assert main_application_window == None
        
        gladefile = os.path.join(config.path.resourcedir, 'pybik.glade')
        #gladefile = os.path.join(resourcepath, 'pybik.ui')
        self.ui_xml = gtk.glade.XML(gladefile)
        #self.ui_xml = gtk.Builder()
        #self.ui_xml.add_from_file(gladefile)
        
        self.window = self.get_object('window1')
        play_toolbar = self.get_object('toolbar1')
        self.statusbar = self.get_object('statusbar1')
        
        pixmapfile = os.path.join(config.path.pixmapdir, 'pybik.png')
        self.window.set_icon_from_file(pixmapfile)
        
        #self.connect_signals(menus)
        self.window.connect("delete-event", gtk.main_quit)

        self.create_menubar()
        self.create_play_toolbar()
        self.create_gl_area(argv)
        self.create_sidepane()
        
        main_application_window = self.window
    
    get_object = lambda self, obj:      self.ui_xml.get_widget(obj)
    #get_object = lambda self, obj:     self.ui_xml.get_object(obj)
    connect_signals = lambda self, cont:    self.ui_xml.signal_autoconnect(cont)
    #connect_signals = lambda self, cont:   self.ui_xml.connect_signals(cont)
    
    #@debug
    def create_menubar(self):
        menubar = self.get_object('menubar1')
        for n in [1,2,3,4]:
            menuitem = self.get_object('menuitem%d'%n)
            #print 'menuitem',menuitem,'menuitem%d'%n
            menubar.remove(menuitem)
        
        menuitem = gtk.MenuItem(_('_Game'))
        menu = gtk.Menu()
        menuitem_ = gtk.ImageMenuItem('gtk-new')
        menuitem_.connect('activate', self.request_new_game, 0)
        menu.append(menuitem_)
        menuitem_ = gtk.SeparatorMenuItem()
        menu.append(menuitem_)
        menuitem_ = gtk.ImageMenuItem('gtk-quit')
        menuitem_.connect('activate', menus.gtk_main_quit)
        menu.append(menuitem_)
        menuitem.set_submenu(menu)
        menubar.append(menuitem)

        menuitem = gtk.MenuItem(_('_Edit'))
        menu = gtk.Menu()
        menuitem_ = gtk.ImageMenuItem('gtk-preferences')
        menuitem_.connect('activate', self.on_preferences)
        menu.append(menuitem_)
        menuitem.set_submenu(menu)
        menubar.append(menuitem)
        
        menuitem = gtk.MenuItem(_('_View'))
        menu = gtk.Menu()
        menuitem_ = gtk.ImageMenuItem('gtk-select-color')
        menuitem_.connect('activate', self.on_colour_select)
        menu.append(menuitem_)
        menuitem_ = gtk.MenuItem(_('_Play Toolbar'))
        menuitem_.connect('activate', toggle_visibility, play_toolbar)
        menu.append(menuitem_)
        menuitem_ = gtk.MenuItem(_('_Status Bar'))
        menuitem_.connect('activate', toggle_visibility, self.statusbar)
        menu.append(menuitem_)
        menuitem.set_submenu(menu)
        menubar.append(menuitem)

        menuitem = gtk.MenuItem(_('_Help'))
        menu = gtk.Menu()
        menuitem_ = gtk.ImageMenuItem('gtk-about')
        menuitem_.connect('activate', menus.about, self.window)
        menu.append(menuitem_)
        menuitem.set_submenu(menu)
        menubar.append(menuitem)
        
        menubar.show_all()

    @debug
    def create_sidepane(self):
        # Get the menu structure asked for by the collection of plugins.
        list_ = plugin_hooks.startup_plugins() or []
        
        treeview = self.get_object('treeview1')
        treestore = self.get_object('treestore1')
        treestore = gtk.TreeStore(gobject.TYPE_STRING, gobject.TYPE_PYOBJECT)
        treeview.set_model(treestore)
        
        column = gtk.TreeViewColumn(_('Name'), gtk.CellRendererText(), text=0)
        column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column.set_fixed_width(200)
        treeview.append_column(column)
        
        # add the items
        tree_iters = {'':None}
        for func_name, func in list_:
            p = func_name.split('/',4)
            for i in xrange(1,5):
                p_i = p[:i]
                p_is = '/'.join(p_i)
                p_i_p = p[:i-1]
                p_is_p = '/'.join(p_i_p)
                name = p_i[-1]
                if p_is not in tree_iters:
                    #print '|'.join((name, p_is, func_name))
                    ti = treestore.append(tree_iters[p_is_p], [name, func if p_is==func_name else None])
                    tree_iters[p_is] = ti
        treeview.connect('row-activated', self.on_run_script)
        treeview.expand_all()
        treeview.show_all()

    @debug
    def on_run_script(self, treeview, path, view_column):
        name, func = treeview.get_model()[path]
        #print name, func
        if callable(func):
            plugin_hooks.run_script(func, self)

    #@debug
    def create_play_toolbar(self):
        rewind = gtk.Action(   "rewind",   _("Rewind"),
        _("Go to the previous mark (or the beginning) of the sequence of moves"),           gtk.STOCK_MEDIA_REWIND)
        previous = gtk.Action( "previous", _("Back"),
        _("Make one step backwards"),        gtk.STOCK_MEDIA_PREVIOUS)
        stop = gtk.Action(     "stop",     _("Stop"),
        _("Stop running the sequence of moves"), gtk.STOCK_MEDIA_STOP)
        mark = gtk.Action(     "mark",     _("Mark"),
        _("Mark the current place in the sequence of moves"),gtk.STOCK_ADD)
        next = gtk.Action(     "next",     _("Forward"),_("Make one step forwards"),        gtk.STOCK_MEDIA_NEXT)
        play = gtk.Action(     "forward",  _("Play"),
        _("Run forward through the sequence of moves"),gtk.STOCK_MEDIA_PLAY)

        play_toolbar.insert (rewind.create_tool_item (),    -1)
        play_toolbar.insert (previous.create_tool_item (),  -1)
        play_toolbar.insert (stop.create_tool_item (),      -1)
        play_toolbar.insert (mark.create_tool_item (),      -1)
        play_toolbar.insert (next.create_tool_item (),      -1)
        play_toolbar.insert (play.create_tool_item (),      -1)

        rewind.connect  ("activate", ui.request_queue_rewind, self)
        previous.connect("activate", ui.request_back, self)
        stop.connect    ("activate", ui.request_stop)
        mark.connect    ("activate", ui.request_mark_move_queue)
        next.connect    ("activate", ui.request_forward, self)
        play.connect    ("activate", ui.request_play, self)

        play_toolbar.show_all()

        global play_button_list
        play_button_list = play_toolbar.get_children ()

        set_toolbar_state (0)


    #@debug
    def create_gl_area(self, argv):
        target = [  ("text/uri-list",  0,  colour_sel.RDRAG_FILELIST),
                    ("application/x-color",  0,  colour_sel.RDRAG_COLOUR),
                ]
        attribs = [ gdkgl.RGBA,
                    gdkgl.RED_SIZE,    1,
                    gdkgl.GREEN_SIZE,  1,
                    gdkgl.BLUE_SIZE,   1,

                    gdkgl.DOUBLEBUFFER,

                    gdkgl.DEPTH_SIZE , 1,

                    gdkgl.ACCUM_RED_SIZE,  1,
                    gdkgl.ACCUM_GREEN_SIZE,  1,
                    gdkgl.ACCUM_BLUE_SIZE,  1,

                    gdkgl.ATTRIB_LIST_NONE
                ]
        
        glarea.glarea_init(argv)
        glconfig = gdkgl.Config (attribs)

        if not glconfig:
            #g_error ("Cannot get a suitable visual")
            raise Exception("Cannot get a suitable visual")

        glxarea = self.get_object('drawingarea1')
        glarea.glwidget = glxarea
        gtkgl.widget_set_gl_capability (glxarea,
                                      glconfig,
                                      None,
                                      True,
                                      gdkgl.RGBA_TYPE)


        glxarea.drag_dest_set (gtk.DEST_DEFAULT_ALL, target, gdk.ACTION_COPY)
        glxarea.connect ("drag_data_received", colour_sel.drag_data_received)

        glxarea.connect ("realize",         glarea.graphics_area_init)
        glxarea.connect ("expose_event",    glarea.expose)
        glxarea.connect ("configure_event", glarea.resize)


        glxarea.add_events (
                               gdk.KEY_PRESS_MASK | gdk.BUTTON_PRESS_MASK
                               | gdk.BUTTON_RELEASE_MASK
                               # | gdk.BUTTON_MOTION_MASK 
                               | gdk.ENTER_NOTIFY_MASK | gdk.LEAVE_NOTIFY_MASK
                               | gdk.VISIBILITY_NOTIFY_MASK
                               | gdk.POINTER_MOTION_MASK )


        glxarea.connect ("key_press_event",     glarea.cube_orientate_keys)
        glxarea.connect ("key_press_event",     self.debug_keys)
        glxarea.connect ("motion_notify_event", glarea.cube_orientate_mouse)
        glxarea.connect ("scroll_event",        glarea.z_rotate)
        glxarea.connect ("button_press_event",  glarea.buttons)
        glxarea.connect ("button_release_event",glarea.buttons)
        glxarea.connect ("button_press_event",  glarea.cube_controls, self)
    
    @debug
    def start_main_loop (self):
        self.window.show()
        gtk.main()

    # Request that the game be restarted
    # If data is non zero,  then all it's data will be reallocated
    @debug
    def request_new_game (self, unused_widget, new_size):
        ui.request_stop (None)
        if ui.is_animating ():
            ui.abort_animation ()
        gobject.idle_add (menus.new_game, new_size, self)

    def on_preferences (self, unused_widget):
        menus.PreferencesDialog(self)
    
    @debug
    def on_colour_select(self, unused_widget):
        colour_sel.open_ColorSelectDialog(self)

    #@debug
    def debug_keys (self, widget, event):
        shifted = 0
        if event.state & gdk.SHIFT_MASK:
            shifted = 1

        if event.keyval == gtk.keysyms.d:
            import code
            code.interact(local=globals())
            return True
        elif event.keyval == gtk.keysyms.t:
            self.on_colour_select(None)
            return True
        elif event.keyval == gtk.keysyms.q:
            menus.gtk_main_quit()
            return True
        elif gtk.keysyms._0 <= event.keyval <= gtk.keysyms._9:
            if not hasattr(self, 'test_int'):
                self.test_int = 0
            context1 = self.statusbar.get_context_id("test1")
            context2 = self.statusbar.get_context_id("test2")
            if event.keyval == gtk.keysyms._1:
                self.test_int += 1
                mesg_id = self.statusbar.push(context1, 'Test %d' % self.test_int)
            elif event.keyval == gtk.keysyms._2:
                mesg_id = self.statusbar.pop(context1)
            elif event.keyval == gtk.keysyms._3:
                self.test_int += 1
                mesg_id = self.statusbar.push(context2, 'Test %d' % self.test_int)
            elif event.keyval == gtk.keysyms._4:
                mesg_id = self.statusbar.pop(context2)
            print '- context',context1,context2,'mesg_id',mesg_id or '-', 'value', self.test_int
            return True
        return True

    #@debug
    def update_statusbar(self):
        mesg = _("Moves: %d / %d") % (
                   ui.move_queue.progress().current,
                   ui.move_queue.progress().total)
        context = self.statusbar.get_context_id("move-count")
        self.statusbar.pop(context)
        self.statusbar.push(context, mesg)

    #@debug
    def declare_win(self, status):
        if status == cube.SOLVED:
            mesg = ngettext("Cube solved in %d move",
                             "Cube solved in %d moves", ui.move_queue.progress().current
                             ) % ui.move_queue.progress().current
        elif status == cube.HALF_SOLVED:
            mesg = _("Cube is NOT solved! The colours are correct,  but have incorrect orientation")
        else:
            assert False, 'not_reached'
        context = self.statusbar.get_context_id("win")
        self.statusbar.pop(context)
        self.statusbar.push(context, mesg)


#@debug
def set_toolbar_state (flags):
    for list_data in play_button_list:
        list_data.set_sensitive (flags & 1)
        flags >>= 1


# Toggle the visibility of a widget
#@debug
def toggle_visibility (menuitem, widget):
    assert widget, '%s, %s' % (menuitem, widget)

    if widget.props.visible:
        widget.hide()
    else:
        widget.show()


# Popup an error dialog box
@debug
def error_dialog (parent, format, *ap):
    message = format % ap
    
    dialog = gtk.MessageDialog (parent,
                     gtk.DIALOG_MODAL,
                     gtk.MESSAGE_ERROR,
                     gtk.BUTTONS_CLOSE,
                     message)
    
    dialog.set_transient_for (parent)
    # Destroy the dialog when the user responds to it (e.g. clicks a button)
    #dialog.connect ("response", gtk.Widget.destroy)
    #dialog.show_all ()
    dialog.run()
    dialog.destroy()


